<?php

namespace Drupal\entity_workflow_workspace\EventSubscriber;

use Drupal\Core\Entity\Sql\SqlContentEntityStorage;
use Drupal\Core\Field\FieldStorageDefinitionEvent;
use Drupal\Core\Field\FieldStorageDefinitionEvents;
use Drupal\entity_workflow\Event\EntityWorkflowEvents;
use Drupal\entity_workflow\Event\InitiateTransitionEvent;
use Drupal\entity_workflow\Event\PostTransitionEvent;
use Drupal\workflows\Entity\Workflow;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class EntityWorkflowWorkspaceEventSubscriber implements EventSubscriberInterface {

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events = [
      // We need a higher priority than
      // \Drupal\entity_workflow\EventSubscriber\EntityWorkflowEventSubscriber
      // so we can publish the workspace *before* initiating the workflow
      // transition.
      EntityWorkflowEvents::INITIATE_TRANSITION => ['onInitiateTransaction', 50],
      PostTransitionEvent::class => ['postTransition', 0],
      FieldStorageDefinitionEvents::CREATE => ['onFieldStorageDefinitionCreate', 150],
    ];
    return $events;
  }

  /**
   * Applies the workflow transition
   *
   * @param \Drupal\entity_workflow\Event\InitiateTransitionEvent $event
   *   The initiate transition event.
   */
  public function onInitiateTransaction(InitiateTransitionEvent $event) {
    if ($event->getWorkflow()->id() === 'workspace') {
      /** @var \Drupal\workspaces\WorkspaceInterface $workspace */
      $workspace = $event->getEntity();

      // Workspace publishing must be executed for any transition that goes to
      // the 'published' state.
      if ($event->getTransition()->to()->id() === 'published') {
        $workspace->publish();
        // @todo When the workspace publisher will invoke a post publish event,
        //   execute the workflow transition there and stop propagating this event.
      }

      // Workspace reverting is tied to our specific 'unpublish' transition ID.
      if ($event->getTransition()->id() === 'unpublish') {
        \Drupal::service('wse.workspace_reverter')->revert($workspace);
      }
    }
  }

  /**
   * @param \Drupal\entity_workflow\Event\PostTransitionEvent $event
   */
  public function postTransition(PostTransitionEvent $event) {
    $workflow = $event->getWorkflow();
    $entity = $event->getEntity();
    $transition = $event->getTransition();
    $context = $event->getContext();

    $is_cascaded_transition = isset($context['cascaded']);
    $is_bulk_transition = isset($context['bulk_transition']);

    // When an entity is transitioned to a new state, ensure that its workspace
    // is automatically transitioned to the "lowest" state of its contents.
    if ($workflow->id() === 'content' && $entity->isDefaultTranslation() && !$is_cascaded_transition && !$is_bulk_transition) {
      /** @var \Drupal\workspaces\WorkspaceInterface|null $workspace */
      $workspace = $entity->workspace->entity;
      if (!$workspace) {
        return;
      }

      // Gather a list of all workflow states for the content of this workspace.
      $all_states = [$transition->to()->id()];

      $tracked_entities = \Drupal::service('workspaces.association')->getTrackedEntities($workspace->id());
      foreach ($tracked_entities as $entity_type_id => $tracked_entity_ids) {
        // The workflow state of the entity that was saved has been added above,
        // so we need to exclude it from the list.
        $exclude_revision_ids = $entity->getEntityTypeId() === $entity_type_id ? [$entity->getRevisionId()] : NULL;

        $states = entity_workflow_get_entity_states('content', $workspace->id(), $entity_type_id, array_keys($tracked_entity_ids), $exclude_revision_ids);
        $all_states = array_unique(array_merge($all_states, $states));
      }
      $all_states = array_filter($all_states);

      // Ensure that states are ordered by weight.
      // @see \Drupal\workflows\Plugin\WorkflowTypeBase::getStates()
      if (count($all_states) > 1) {
        $all_states = array_keys(Workflow::load('content')->getTypePlugin()->getStates($all_states));
      }
      $lowest_content_state = reset($all_states);

      // Transition the workspace to the new "lowest content state" if needed.
      $workflow_field = entity_workflow_get_field($workspace, 'workspace');
      /** @var \Drupal\entity_workflow_workspace\Plugin\WorkflowType\EntityWorkflowWorkspace $workspace_workflow */
      $workspace_workflow = $workflow_field->getWorkflow()->getTypePlugin();

      if ($workflow_field->getStateId() != $lowest_content_state && $workspace_workflow->hasState($lowest_content_state)) {
        entity_workflow_system_transition($workspace, 'workspace', $lowest_content_state, '', ['cascaded' => TRUE]);
      }
    }
  }

  /**
   * Reacts to the creation of a field storage definition.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionEvent $event
   *   The event.
   */
  public function onFieldStorageDefinitionCreate(FieldStorageDefinitionEvent $event): void {
    if ($event->getFieldStorageDefinition()->getName() !== entity_workflow_get_field_name('workspace')) {
      return;
    }

    $storage = \Drupal::entityTypeManager()->getStorage('workspace');
    if (!($storage instanceof SqlContentEntityStorage)) {
      return;
    }

    /** @var \Drupal\Core\Entity\EntityLastInstalledSchemaRepositoryInterface $last_installed_schema_repository */
    $last_installed_schema_repository = \Drupal::service('entity.last_installed_schema.repository');
    $storage_definitions = $last_installed_schema_repository->getLastInstalledFieldStorageDefinitions('workspace');

    if (isset($storage_definitions['status']) && $storage_definitions['status']->getProvider() === 'wse') {
      $open_workspace_ids = $storage->getQuery()
        ->accessCheck(FALSE)
        ->condition('status', WSE_STATUS_OPEN)
        ->execute();
      $closed_workspace_ids = $storage->getQuery()
        ->accessCheck(FALSE)
        ->condition('status', WSE_STATUS_CLOSED)
        ->execute();

      $database = \Drupal::database();
      $table_mapping = $storage->getTableMapping();
      $workflow_column = $table_mapping->getFieldColumnName($storage_definitions['entity_workflow_workspace'], 'value');
      $id_column = $table_mapping->getFieldColumnName($storage_definitions['id'], 'value');
      foreach ($table_mapping->getAllFieldTableNames('entity_workflow_workspace') as $table_name) {
        if ($open_workspace_ids) {
          $database->update($table_name)
            ->fields([$workflow_column => 'draft'])
            ->condition($id_column, array_values($open_workspace_ids), 'IN')
            ->execute();
        }
        if ($closed_workspace_ids) {
          $database->update($table_name)
            ->fields([$workflow_column => 'published'])
            ->condition($id_column, array_values($closed_workspace_ids), 'IN')
            ->execute();
        }
      }

      $storage->resetCache();

      // Rebuild the router so workflow actions links are correctly assigned.
      \Drupal::service('router.builder')->setRebuildNeeded();
    }
  }

}
