<?php

namespace Drupal\entity_workflow\Form;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Utility\Error;
use Drupal\entity_workflow\EntityWorkflowInfo;
use Drupal\entity_workflow\EntityWorkflowTransition;
use Drupal\entity_workflow\StateTransitionValidatorInterface;
use Drupal\workflows\WorkflowInterface;
use Drupal\workspaces\WorkspaceManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class EntityWorkflowForm extends FormBase implements EntityWorkflowFormInterface {

  /**
   * The entity workflow information service.
   *
   * @var \Drupal\entity_workflow\EntityWorkflowInfo
   */
  protected $entityWorkflowInfo;

  /**
   * The state transition validation service.
   *
   * @var \Drupal\entity_workflow\StateTransitionValidatorInterface
   */
  protected $stateTransitionValidator;

  /**
   * The workspace manager.
   *
   * @var \Drupal\workspaces\WorkspaceManagerInterface
   */
  protected $workspaceManager;

  /**
   * @var \Drupal\Core\Entity\ContentEntityInterface
   */
  protected $entity;

  /**
   * @var \Drupal\workflows\WorkflowInterface
   */
  protected $workflow;

  /**
   * @var \Drupal\entity_workflow\WorkflowType\EntityWorkflowTypeInterface
   */
  protected $workflowPlugin;

  /**
   * The transition that will be applied.
   */
  protected ?EntityWorkflowTransition $transition;

  /**
   * Creates a EntityWorkflowForm object.
   *
   * @param \Drupal\entity_workflow\EntityWorkflowInfo $entity_workflow_info
   *   The entity workflow information service.
   * @param \Drupal\entity_workflow\StateTransitionValidatorInterface $state_transition_validation
   *   The state transition validation service.
   * @param \Drupal\workspaces\WorkspaceManagerInterface $workspace_manager
   *   The workspace manager.
   */
  public function __construct(EntityWorkflowInfo $entity_workflow_info, StateTransitionValidatorInterface $state_transition_validation, WorkspaceManagerInterface $workspace_manager) {
    $this->entityWorkflowInfo = $entity_workflow_info;
    $this->stateTransitionValidator = $state_transition_validation;
    $this->workspaceManager = $workspace_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_workflow.info'),
      $container->get('entity_workflow.state_transition_validation'),
      $container->get('workspaces.manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'entity_workflow_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, WorkflowInterface $workflow = NULL, ContentEntityInterface $entity = NULL) {
    $this->entity = $entity;
    $this->workflow = $workflow;
    $this->workflowPlugin = $this->workflow->getTypePlugin();

    $workspace = $this->entity->getEntityTypeId() === 'workspace' ? $this->entity : $this->workspaceManager->getActiveWorkspace();
    // @todo Implement workspace states.
    // Do not display the form at all if the workspace is in a closed state.
    //$states = cps_changeset_get_states();
    //if ($states[$changeset->status]['type'] != 'open') {
    //  return $form;
    //}

    $form['#title'] = $this->t('@label workflow', ['@label' => $this->workflow->label()]);

    $state = entity_workflow_get_entity_state($this->entity, $this->workflow->id());
    $form['state'] = [
      '#type' => 'item',
      '#title' => $this->t('Current state'),
      '#markup' => $this->workflowPlugin->getState($state)->label(),
      '#wrapper_attributes' => ['class' => ['container-inline']],
    ];

    $available_transitions = $this->stateTransitionValidator->getValidTransitions($this->entity, $this->workflow, $this->currentUser(), $workspace);
    $available_transition_options = array_map(function (EntityWorkflowTransition $transition) {
      return $transition->label();
    }, $available_transitions);

    $form['transition'] = [
      '#type' => 'radios',
      '#options' => $available_transition_options,
      '#required' => TRUE,
    ];

    // Create a list of transitions that use log messages so we can
    // apply #states and properly validate. Exclude unavailable transitions
    // since these will never need a log message.
    $log_transitions = [];
    foreach ($available_transitions as $transition_id => $transition) {
      if ($transition->hasLog()) {
        $log_transitions[] = $transition_id;
        if ($transition->isLogRequired()) {
          $log_required[$transition_id] = $transition_id;
        }
      }
    }
    if (!empty($log_transitions)) {
      $form['log'] = [
        '#type' => 'textarea',
        '#title' => $this->t('Provide a log message'),
        '#rows' => 4,
        '#weight' => 10,
      ];

      foreach ($log_transitions as $transition_id) {
        $form['log']['#states']['visible'][':input[name="transition"]'][] = ['value' => $transition_id];
      }

      if (!empty($log_required)) {
        foreach ($log_required as $transition_id) {
          $form['log']['#states']['required'][':input[name="transition"]'][] = ['value' => $transition_id];
        }
      }
    }

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Change'),
      '#button_type' => 'primary',
      // Do not show the button if there are no transitions.
      '#access' => !empty($available_transitions),
    ];

    $form['#attributes']['class'][] = 'entity-workflow-entity-state-form';

    // Changing the workflow state of an entity is allowed in a workspace.
    $form_state->set('workspace_safe', TRUE);

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    $transition = $this->workflowPlugin->getTransition($form_state->getValue('transition'));
    $this->transition = $transition;

    if ($transition->hasLog() && $transition->isLogRequired() && empty($form_state->getValue('log'))) {
      $form_state->setError($form['log'], $this->t('Log field is required'));
    }

    // If they entered a log message when it was visible and then switched
    // to something where the log message was not visible, remove it so it
    // does not get recorded.
    if (!$transition->hasLog()) {
      $form_state->setValue('log', '');
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $log = $form_state->getValue('log');
    $workspace = $this->workspaceManager->hasActiveWorkspace() ? $this->workspaceManager->getActiveWorkspace() : NULL;

    try {
      if ($this->stateTransitionValidator->isTransitionValid($this->entity, $this->workflow, $this->transition->id(), $this->currentUser(), $workspace)) {
        entity_workflow_transition($this->entity, $this->workflow->id(), $this->transition->id(), $log);
        $this->messenger()->addMessage($this->t('The workflow state has been changed.'));
      }
      else {
        $this->messenger()->addError($this->t('The workflow state could not be changed.'));
      }
    }
    catch (\Exception $e) {
      Error::logException($this->logger('entity_workflow'), $e);
      $this->messenger()->addError($this->t('An error occurred while changing the workflow state. See logs for more details.'));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity(): EntityInterface {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getTransition(): ?EntityWorkflowTransition {
    return $this->transition;
  }

}
